@@ -30,27 +30,6 @@
 #include <signal.h>
 #include <errno.h>
 
-#if defined (HAVE_IEEEFP_H) || defined (HAVE_IEEE754_H)
-/* Make sure we have this symbol defined, since the existance of either
-   header file implies it.  */
-#ifndef IEEE_754
-#define IEEE_754
-#endif
-#endif
-
-#ifdef HAVE_SUNMATH_H
-#include <sunmath.h>
-#endif
-
-#ifdef HAVE_IEEEFP_H
-#include <ieeefp.h>
-#endif
-#ifdef HAVE_IEEE754_H
-#include <ieee754.h>
-#endif
-
-#define ML_UNDERFLOW (GO_EPSILON * GO_EPSILON)
-
 double go_nan;
 double go_pinf;
 double go_ninf;
@@ -64,7 +43,7 @@ running_under_buggy_valgrind (void)
 		return FALSE;
 
 	/*
-	 * We get here is long double fails to satisfy a requirement of
+	 * We get here when long double fails to satisfy a requirement of
 	 * C99, namely that LDBL_MIN is positive.  That is probably
 	 * valgrind mapping long doubles to doubles.
 	 *
@@ -212,6 +191,9 @@ _go_math_init (void)
 double
 go_add_epsilon (double x)
 {
+#ifdef HAVE_NEXTAFTER
+	return x == 0 ? x : nextafter (x, go_pinf);
+#else
 	if (!go_finite (x) || x == 0)
 		return x;
 	else {
@@ -220,11 +202,15 @@ go_add_epsilon (double x)
 		double absres = ldexp (mant + DBL_EPSILON, exp);
 		return (x < 0) ? -absres : absres;
 	}
+#endif
 }
 
 double
 go_sub_epsilon (double x)
 {
+#ifdef HAVE_NEXTAFTER
+	return x == 0 ? x : nextafter (x, go_ninf);
+#else
 	if (!go_finite (x) || x == 0)
 		return x;
 	else {
@@ -233,6 +219,7 @@ go_sub_epsilon (double x)
 		double absres = ldexp (mant - DBL_EPSILON, exp);
 		return (x < 0) ? -absres : absres;
 	}
+#endif
 }
 
 static double
@@ -445,22 +432,22 @@ go_ascii_strtod (const char *s, char **end)
 double
 log1p (double x)
 {
-  double term, sum;
-  int i;
+	double term, sum;
+	int i;
 
-  if (fabs (x) > 0.25)
-    return log (x + 1);
+	if (fabs (x) > 0.25)
+		return log (x + 1);
 
-  i = 0;
-  sum = 0;
-  term = -1;
-  while (fabs (term) > fabs (sum) * DBL_EPSILON) {
-    term *= -x;
-    i++;
-    sum += term / i;
-  }
+	i = 0;
+	sum = 0;
+	term = -1;
+	while (fabs (term) > fabs (sum) * DBL_EPSILON) {
+		term *= -x;
+		i++;
+		sum += term / i;
+	}
 
-  return sum;
+	return sum;
 }
 #endif
 
@@ -491,9 +478,9 @@ expm1 (double x)
 double
 asinh (double x)
 {
-  double y = fabs (x);
-  double r = log1p (y * y / (hypot (y, 1.0) + 1.0) + y);
-  return (x >= 0) ? r : -r;
+	double y = fabs (x);
+	double r = log1p (y * y / (hypot (y, 1.0) + 1.0) + y);
+	return (x >= 0) ? r : -r;
 }
 #endif
 
@@ -501,8 +488,8 @@ asinh (double x)
 double
 acosh (double x)
 {
-  double xm1 = x - 1;
-  return log1p (xm1 + sqrt (xm1) * sqrt (x + 1.0));
+	double xm1 = x - 1;
+	return log1p (xm1 + sqrt (xm1) * sqrt (x + 1.0));
 }
 #endif
 
@@ -510,9 +497,9 @@ acosh (double x)
 double
 atanh (double x)
 {
-  double y = fabs (x);
-  double r = -0.5 * log1p (-(y + y) / (1.0 + y));
-  return (x >= 0) ? r : -r;
+	double y = fabs (x);
+	double r = -0.5 * log1p (-(y + y) / (1.0 + y));
+	return (x >= 0) ? r : -r;
 }
 #endif
 
@@ -648,6 +635,9 @@ go_ascii_strtold (const char *s, char **end)
 long double
 go_add_epsilonl (long double x)
 {
+#ifdef HAVE_NEXTAFTERL
+	return x == 0 ? x : nextafterl (x, go_pinfl);
+#else
 	if (!go_finitel (x) || x == 0)
 		return x;
 	else {
@@ -656,11 +646,15 @@ go_add_epsilonl (long double x)
 		long double absres = ldexpl (mant + LDBL_EPSILON, exp);
 		return (x < 0) ? -absres : absres;
 	}
+#endif
 }
 
 long double
 go_sub_epsilonl (long double x)
 {
+#ifdef HAVE_NEXTAFTERL
+	return x == 0 ? x : nextafterl (x, go_ninfl);
+#else
 	if (!go_finitel (x) || x == 0)
 		return x;
 	else {
@@ -669,6 +663,7 @@ go_sub_epsilonl (long double x)
 		long double absres = ldexpl (mant - LDBL_EPSILON, exp);
 		return (x < 0) ? -absres : absres;
 	}
+#endif
 }
 
 long double
@@ -907,3 +902,219 @@ go_stern_brocot (double val, int max_denom, int *res_num, int *res_denom)
 		*res_denom = bd;
 	}
 }
+
+/* ------------------------------------------------------------------------- */
+
+static double
+reduce_half (double x, int *pk)
+{
+	int k = 0;
+
+	if (x < 0) {
+		x = -reduce_half (-x, &k);
+		k = 4 - k;
+		if (x == -0.25)
+			x += 0.5, k += 3;
+	} else {
+		x = fmod (x, 2);
+		if (x >= 1)
+			x -= 1, k += 2;
+		if (x >= 0.5)
+			x -= 0.5, k++;
+		if (x > 0.25)
+			x -= 0.5, k++;
+	}
+
+	*pk = (k & 3);
+	return x;
+}
+
+static double
+do_sinpi (double x, int k)
+{
+	double y;
+
+	if (x == 0)
+		y = k & 1;
+	else if (x == 0.25)
+		y = 0.707106781186547524400844362104849039284835937688474036588339;
+	else
+		y = (k & 1) ? cos (M_PI * x) : sin (M_PI * x);
+
+	return (k & 2) ? 0 - y : y;
+}
+
+
+/**
+ * go_sinpi:
+ * @x: a number
+ *
+ * Returns: the sine of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+double
+go_sinpi (double x)
+{
+	int k;
+	x = reduce_half (x, &k);
+	return do_sinpi (x, k);
+}
+
+/**
+ * go_cospi:
+ * @x: a number
+ *
+ * Returns: the cosine of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+double
+go_cospi (double x)
+{
+	int k;
+	x = reduce_half (x, &k);
+	return do_sinpi (x, k + 1);
+}
+
+/**
+ * go_tanpi:
+ * @x: a number
+ *
+ * Returns: the tangent of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+double
+go_tanpi (double x)
+{
+	int k;
+	x = reduce_half (x, &k);
+	return do_sinpi (x, k) / do_sinpi (x, k + 1);
+}
+
+double
+go_atan2pi (double y, double x)
+{
+	/* This ignores loads of cases with -0. */
+
+	if (y == 0)
+		return (x >= 0 ? 0 : +1);
+
+	if (x == 0)
+		return (y >= 0 ? 0.5 : -0.5);
+
+	if (fabs (x) == fabs (y))
+		return (y > 0 ? +1 : -1) * (x > 0 ? 0.25 : 0.75);
+
+	/* Fallback.  */
+	return atan2 (y, x) / M_PI;
+}
+
+#ifdef GOFFICE_WITH_LONG_DOUBLE
+
+#define M_PIgo    3.141592653589793238462643383279502884197169399375105820974944592307816406286208998628034825342117L
+
+static long double
+reduce_halfl (long double x, int *pk)
+{
+	int k = 0;
+
+	if (x < 0) {
+		x = -reduce_halfl (-x, &k);
+		k = 4 - k;
+		if (x == -0.25)
+			x += 0.5, k += 3;
+	} else {
+		x = fmod (x, 2);
+		if (x >= 1)
+			x -= 1, k += 2;
+		if (x >= 0.5)
+			x -= 0.5, k++;
+		if (x > 0.25)
+			x -= 0.5, k++;
+	}
+
+	*pk = (k & 3);
+	return x;
+}
+
+static long double
+do_sinpil (long double x, int k)
+{
+	long double y;
+
+	if (x == 0)
+		y = k & 1;
+	else if (x == 0.25)
+		y = 0.707106781186547524400844362104849039284835937688474036588339L;
+	else
+		y = (k & 1) ? cos (M_PIgo * x) : sin (M_PIgo * x);
+
+	return (k & 2) ? 0 - y : y;
+}
+
+
+/**
+ * go_sinpil:
+ * @x: a number
+ *
+ * Returns: the sine of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+long double
+go_sinpil (long double x)
+{
+	int k;
+	x = reduce_halfl (x, &k);
+	return do_sinpil (x, k);
+}
+
+/**
+ * go_cospil:
+ * @x: a number
+ *
+ * Returns: the cosine of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+long double
+go_cospil (long double x)
+{
+	int k;
+	x = reduce_halfl (x, &k);
+	return do_sinpil (x, k + 1);
+}
+
+/**
+ * go_tanpil:
+ * @x: a number
+ *
+ * Returns: the tangent of Pi times @x, but with less error than doing the
+ * multiplication outright.
+ */
+long double
+go_tanpil (long double x)
+{
+	int k;
+	x = reduce_halfl (x, &k);
+	return do_sinpil (x, k) / do_sinpil (x, k + 1);
+}
+
+long double
+go_atan2pil (long double y, long double x)
+{
+	/* This ignores loads of cases with -0. */
+
+	if (y == 0)
+		return (x >= 0 ? 0 : +1);
+
+	if (x == 0)
+		return (y >= 0 ? 0.5 : -0.5);
+
+	if (fabs (x) == fabs (y))
+		return (y > 0 ? +1 : -1) * (x > 0 ? 0.25 : 0.75);
+
+	/* Fallback.  */
+	return atan2l (y, x) / M_PIgo;
+}
+
+#endif
+
+/* ------------------------------------------------------------------------- */
