@@ -215,6 +215,12 @@ SUFFIX(go_quad_start) (void)
 					   256.0,
 					   256.0);
 
+		SUFFIX(go_quad_constant8) ((QUAD *)&SUFFIX(go_quad_2pi),
+					   pi_hex_digits,
+					   G_N_ELEMENTS (pi_hex_digits),
+					   256.0,
+					   512.0);
+
 		SUFFIX(go_quad_constant8) ((QUAD *)&SUFFIX(go_quad_e),
 					   e_hex_digits,
 					   G_N_ELEMENTS (e_hex_digits),
@@ -265,9 +271,10 @@ SUFFIX(go_quad_end) (void *state)
 	g_free (state);
 }
 
-const QUAD SUFFIX(go_quad_zero);
+const QUAD SUFFIX(go_quad_zero) = { 0, 0 };
 const QUAD SUFFIX(go_quad_one) = { 1, 0 };
 const QUAD SUFFIX(go_quad_pi);
+const QUAD SUFFIX(go_quad_2pi);
 const QUAD SUFFIX(go_quad_e);
 const QUAD SUFFIX(go_quad_ln2);
 const QUAD SUFFIX(go_quad_sqrt2);
@@ -611,7 +618,7 @@ SUFFIX(go_quad_pow_frac) (QUAD *res, const QUAD *x, const QUAD *y,
 /*
  * This computes pow(x,y) in the following way:
  *
- * 1. y is considered the sum of a number of one-but values.
+ * 1. y is considered the sum of a number of one-bit values.
  * 2. For each y bit in the integer part of y, the corresponding x^y
  *    is computed by repeated squaring.
  * 3. For each y bit in the fractional part of y, the corresponding x^y
@@ -734,3 +741,539 @@ SUFFIX(go_quad_expm1) (QUAD *res, const QUAD *a)
 		SUFFIX(go_quad_div) (res, &z, &zp1);
 	}
 }
+
+/**
+ * go_quad_log: (skip)
+ **/
+/**
+ * go_quad_logl: (skip)
+ **/
+void
+SUFFIX(go_quad_log) (QUAD *res, const QUAD *a)
+{
+	DOUBLE da = SUFFIX(go_quad_value) (a);
+
+	if (da == 0)
+		SUFFIX(go_quad_init) (res, SUFFIX(go_ninf));
+	else if (da < 0)
+		SUFFIX(go_quad_init) (res, SUFFIX(go_nan));
+	else if (!SUFFIX(go_finite) (da))
+		*res = *a;
+	else {
+		QUAD xi, yi, dx;
+		SUFFIX(go_quad_init) (&xi, SUFFIX(log) (da));
+
+		/* Newton step. */
+		SUFFIX(go_quad_exp) (&yi, NULL, &xi);
+		SUFFIX(go_quad_sub) (&dx, a, &yi);
+		SUFFIX(go_quad_div) (&dx, &dx, &yi);
+		SUFFIX(go_quad_add) (&xi, &xi, &dx);
+		
+		*res = xi;
+	}
+}
+
+void
+SUFFIX(go_quad_hypot) (QUAD *res, const QUAD *a, const QUAD *b)
+{
+	int e;
+	QUAD qa2, qb2, qn;
+
+	if (a->h == 0) {
+		res->h = SUFFIX(fabs)(b->h);
+		res->l = SUFFIX(fabs)(b->l);
+		return;
+	}
+	if (b->h == 0) {
+		res->h = SUFFIX(fabs)(a->h);
+		res->l = SUFFIX(fabs)(a->l);
+		return;
+	}
+
+	/* Scale by power of 2 to protect against over- and underflow */
+	(void)SUFFIX(frexp) (MAX (SUFFIX(fabs) (a->h), SUFFIX(fabs) (b->h)), &e);
+
+	qa2.h = SUFFIX(ldexp) (a->h, -e);
+	qa2.l = SUFFIX(ldexp) (a->l, -e);
+	SUFFIX(go_quad_mul) (&qa2, &qa2, &qa2);
+
+	qb2.h = SUFFIX(ldexp) (b->h, -e);
+	qb2.l = SUFFIX(ldexp) (b->l, -e);
+	SUFFIX(go_quad_mul) (&qb2, &qb2, &qb2);
+
+	SUFFIX(go_quad_add) (&qn, &qa2, &qb2);
+	SUFFIX(go_quad_sqrt) (&qn, &qn);
+	res->h = SUFFIX(ldexp) (qn.h, e);
+	res->l = SUFFIX(ldexp) (qn.l, e);
+}
+
+/* sqrt(1-a*a) helper */
+static void
+SUFFIX(go_quad_ihypot) (QUAD *res, const QUAD *a)
+{
+	QUAD qp;
+
+	SUFFIX(go_quad_mul) (&qp, a, a);
+	SUFFIX(go_quad_sub) (&qp, &SUFFIX(go_quad_one), &qp);
+	SUFFIX(go_quad_sqrt) (res, &qp);
+}
+
+
+#ifdef DEFINE_COMMON
+typedef enum {
+	AGM_ARCSIN,
+	AGM_ARCCOS,
+	AGM_ARCTAN
+} AGM_Method;
+#endif
+
+static void
+SUFFIX(go_quad_agm_internal) (QUAD *res, AGM_Method method, const QUAD *x)
+{
+	QUAD g, gp, dk[20], dpk[20], qr, qrp, qnum;
+	int n, k;
+	gboolean converged = FALSE;
+
+	/*
+	 * This follows "An Algorithm for Computing Logarithms
+	 * and Arctangents" by B. C. Carlson in *Mathematics of
+	 * Computation*, Volume 26, Number 118, April 1972.
+	 *
+	 * If need be we can do log, arctanh, arcsinh, and
+	 * arccosh we the same code.
+	 */
+
+	qrp = SUFFIX(go_quad_zero);
+
+	switch (method) {
+	case AGM_ARCSIN:
+		g_return_if_fail (SUFFIX(fabs) (x->h) <= 1);
+		SUFFIX(go_quad_ihypot) (&dpk[0], x);
+		gp = SUFFIX(go_quad_one);
+		qnum = *x;
+		break;
+	case AGM_ARCCOS:
+		g_return_if_fail (SUFFIX(fabs) (x->h) <= 1);
+		dpk[0] = *x;
+		gp = SUFFIX(go_quad_one);
+		SUFFIX(go_quad_ihypot) (&qnum, x);
+		break;
+	case AGM_ARCTAN:
+		g_return_if_fail (SUFFIX(fabs) (x->h) <= 1);
+		dpk[0] = SUFFIX(go_quad_one);
+		SUFFIX(go_quad_hypot) (&gp, x, &SUFFIX(go_quad_one));
+		qnum = *x;
+		break;
+	default:
+		g_assert_not_reached ();
+	}
+
+	for (n = 1; n < (int)G_N_ELEMENTS(dk); n++) {
+		SUFFIX(go_quad_add) (&dk[0], &dpk[0], &gp);
+		dk[0].h *= 0.5; dk[0].l *= 0.5;
+
+		SUFFIX(go_quad_mul) (&g, &dk[0], &gp);
+		SUFFIX(go_quad_sqrt) (&g, &g);
+
+		for (k = 1; k <= n; k++) {
+			QUAD f;
+
+			SUFFIX(go_quad_init) (&f, SUFFIX(ldexp) (1, -(2 * k)));
+			SUFFIX(go_quad_mul) (&dk[k], &f, &dpk[k-1]);
+			SUFFIX(go_quad_sub) (&dk[k], &dk[k-1], &dk[k]);
+			SUFFIX(go_quad_sub) (&f, &SUFFIX(go_quad_one), &f);
+			SUFFIX(go_quad_div) (&dk[k], &dk[k], &f);
+		}
+
+		SUFFIX(go_quad_div) (&qr, &qnum, &dk[n]);
+		SUFFIX(go_quad_sub) (&qrp, &qrp, &qr);
+		if (SUFFIX(fabs)(qrp.h) <= SUFFIX(ldexp) (SUFFIX(fabs)(qr.h), -2 * (DOUBLE_MANT_DIG - 1))) {
+			converged = TRUE;
+			break;
+		}
+
+		qrp = qr;
+		gp = g;
+		for (k = 0; k <= n; k++) dpk[k] = dk[k];
+	}
+
+	if (!converged)
+		g_warning ("go_quad_agm_internal(%.20g) failed to converge\n",
+			   (double)SUFFIX(go_quad_value) (x));
+
+	*res = qr;
+}
+
+static gboolean
+SUFFIX(go_quad_atan2_special) (const QUAD *y, const QUAD *x, DOUBLE *f)
+{
+	DOUBLE dy = SUFFIX(go_quad_value) (y);
+	DOUBLE dx = SUFFIX(go_quad_value) (x);
+
+	if (dy == 0) {
+		/* This assumes all zeros are +0. */
+		*f = (dx >= 0 ? 0 : +1);
+		return TRUE;
+	}
+
+	if (dx == 0) {
+		*f = (dy >= 0 ? 0.5 : -0.5);
+		return TRUE;
+	}
+
+	if (SUFFIX(fabs) (SUFFIX(fabs)(dx) - SUFFIX(fabs)(dy)) < 1e-10) {
+		QUAD d;
+		SUFFIX(go_quad_sub) (&d, x, y);
+		if (d.h == 0) {
+			*f = (dy >= 0 ? 0.25 : -0.75);
+			return TRUE;
+		}
+		SUFFIX(go_quad_add) (&d, x, y);
+		if (d.h == 0) {
+			*f = (dy >= 0 ? +0.75 : -0.25);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+/**
+ * go_quad_atan2: (skip)
+ **/
+/**
+ * go_quad_atan2l: (skip)
+ **/
+void
+SUFFIX(go_quad_atan2) (QUAD *res, const QUAD *y, const QUAD *x)
+{
+	DOUBLE f;
+	DOUBLE dy = SUFFIX(go_quad_value) (y);
+	DOUBLE dx = SUFFIX(go_quad_value) (x);
+	QUAD qr;
+
+	if (SUFFIX(go_quad_atan2_special) (y, x, &f)) {
+		res->h = f * SUFFIX(go_quad_pi).h;
+		res->l = f * SUFFIX(go_quad_pi).l;
+		return;
+	}
+
+	if (SUFFIX(fabs) (dx) >= SUFFIX(fabs) (dy)) {
+		SUFFIX(go_quad_div) (&qr, y, x);
+		SUFFIX(go_quad_agm_internal) (res, AGM_ARCTAN, &qr);
+	} else {
+		DOUBLE f;
+		QUAD qa;
+
+		SUFFIX(go_quad_div) (&qr, x, y);
+		SUFFIX(go_quad_agm_internal) (res, AGM_ARCTAN, &qr);
+
+		f = (qr.h >= 0) ? 0.5 : -0.5;
+		qa.h = f * SUFFIX(go_quad_pi).h;
+		qa.l = f * SUFFIX(go_quad_pi).l;
+		SUFFIX(go_quad_sub) (res, &qa, res);
+	}
+
+	if (dx < 0) {
+		/* Correct for quadrants 2 and 3. */
+		if (dy > 0)
+			SUFFIX(go_quad_add) (res, res, &SUFFIX(go_quad_pi));
+		else
+			SUFFIX(go_quad_sub) (res, res, &SUFFIX(go_quad_pi));
+	}
+}
+
+/**
+ * go_quad_atan2pi: (skip)
+ **/
+/**
+ * go_quad_atan2pil: (skip)
+ **/
+void
+SUFFIX(go_quad_atan2pi) (QUAD *res, const QUAD *y, const QUAD *x)
+{
+	DOUBLE f;
+
+	if (SUFFIX(go_quad_atan2_special) (y, x, &f)) {
+		SUFFIX(go_quad_init) (res, f);
+		return;
+	}
+
+	/* Fallback.  */
+	SUFFIX(go_quad_atan2) (res, y, x);
+	SUFFIX(go_quad_div) (res, res, &SUFFIX(go_quad_pi));
+}
+
+static gboolean
+SUFFIX(reduce_pi_half) (QUAD *res, const QUAD *a, int *pk)
+{
+	static QUAD pi_half;
+	QUAD qa, qk, qh, qb;
+	DOUBLE k;
+	unsigned ui;
+	static const DOUBLE pi_half_parts[] = {
+		+0x1.921fb54442d18p+0,
+		+0x1.1a62633145c04p-54,
+		+0x1.707344a40938p-105,
+		+0x1.114cf98e80414p-156,
+		+0x1.bea63b139b224p-207,
+		+0x1.14a08798e3404p-259,
+		+0x1.bbdf2a33679a4p-312,
+		+0x1.a431b302b0a6cp-363,
+		+0x1.f25f14374fe1p-415,
+		+0x1.ab6b6a8e122fp-466
+	};
+
+	if (!SUFFIX(go_finite) (a->h))
+		return TRUE;
+
+	if (SUFFIX(fabs) (a->h) > SUFFIX(ldexp) (1.0, DOUBLE_MANT_DIG)) {
+		g_warning ("Reduced accuracy for very large trigonometric arguments");
+		return TRUE;
+	}
+
+	if (pi_half.h == 0) {
+		pi_half.h = SUFFIX(go_quad_pi).h * 0.5;
+		pi_half.l = SUFFIX(go_quad_pi).l * 0.5;
+	}
+
+	SUFFIX(go_quad_div) (&qk, a, &pi_half);
+	qh.h = 0.5; qh.l = 0;
+	SUFFIX(go_quad_add) (&qk, &qk, &qh);
+	SUFFIX(go_quad_floor) (&qk, &qk);
+	k = SUFFIX(go_quad_value) (&qk);
+	*pk = (int)(SUFFIX(fmod) (k, 4));
+
+	qa = *a;
+	for (ui = 0; ui < G_N_ELEMENTS(pi_half_parts); ui++) {
+		SUFFIX(go_quad_mul12) (&qb, pi_half_parts[ui], k);
+		SUFFIX(go_quad_sub) (&qa, &qa, &qb);
+	}
+
+	*res = qa;
+
+	return FALSE;
+}
+
+static void
+SUFFIX(reduce_half) (QUAD *res, const QUAD *a, int *pk)
+{
+	static const QUAD half = { 0.5, 0 };
+	int k = 0;
+	QUAD qxr = *a;
+
+	if (a->h < 0) {
+		QUAD aa;
+		aa.h = -a->h; aa.l = -a->l;
+		SUFFIX(reduce_half) (&qxr, &aa, &k);
+		qxr.h = -qxr.h; qxr.l = -qxr.l;
+		k = 4 - k;
+		if (qxr.h <= -0.25 && qxr.l == 0) {
+			SUFFIX(go_quad_add) (&qxr, &qxr, &half);
+			k += 3;
+		}
+	} else {
+		QUAD qdx;
+
+		/* Remove integer multiples of 2.  */
+		SUFFIX(go_quad_init) (&qdx, qxr.h - SUFFIX(fmod) (qxr.h, 2));
+		SUFFIX(go_quad_sub) (&qxr, &qxr, &qdx);
+
+		/* Again, just in case it was really big.  */
+		SUFFIX(go_quad_init) (&qdx, qxr.h - SUFFIX(fmod) (qxr.h, 2));
+		SUFFIX(go_quad_sub) (&qxr, &qxr, &qdx);
+
+		if (qxr.h >= 1) {
+			SUFFIX(go_quad_sub) (&qxr, &qxr, &SUFFIX(go_quad_one));
+			k += 2;
+		}
+		if (qxr.h >= 0.5) {
+			SUFFIX(go_quad_sub) (&qxr, &qxr, &half);
+			k++;
+		}
+		if (qxr.h > 0.25) {
+			SUFFIX(go_quad_sub) (&qxr, &qxr, &half);
+			k++;
+		}
+	}
+
+	*pk = (k & 3);
+	*res = qxr;
+}
+
+static void
+SUFFIX(do_sin) (QUAD *res, const QUAD *a, int k)
+{
+	QUAD qr;
+
+	if (k & 1) {
+		QUAD qn, qd, qq, aa;
+
+		aa.h = SUFFIX(fabs)(a->h);
+		aa.l = SUFFIX(fabs)(a->l);
+		SUFFIX(go_quad_init) (&qr, cos (aa.h));
+
+		/* Newton step */
+		SUFFIX(go_quad_acos) (&qn, &qr);
+		SUFFIX(go_quad_sub) (&qn, &qn, &aa);
+		SUFFIX(go_quad_ihypot) (&qd, &qr);
+		SUFFIX(go_quad_mul) (&qq, &qn, &qd);
+		SUFFIX(go_quad_add) (&qr, &qr, &qq);
+	} else {
+		QUAD qn, qd, qq;
+		SUFFIX(go_quad_init) (&qr, sin (a->h));
+
+		/* Newton step */
+		SUFFIX(go_quad_asin) (&qn, &qr);
+		SUFFIX(go_quad_sub) (&qn, &qn, a);
+		SUFFIX(go_quad_ihypot) (&qd, &qr);
+		SUFFIX(go_quad_mul) (&qq, &qn, &qd);
+		SUFFIX(go_quad_sub) (&qr, &qr, &qq);
+	}
+
+	if (k & 2) {
+		qr.h = 0 - qr.h;
+		qr.l = 0 - qr.l;
+	}
+
+	*res = qr;
+}
+
+static void
+SUFFIX(do_sinpi) (QUAD *res, const QUAD *a, int k)
+{
+	QUAD qr;
+
+	if (a->h == 0)
+		SUFFIX(go_quad_init) (&qr, k & 1);
+	else if (a->h == 0.25 && a->l == 0)
+		SUFFIX(go_quad_div) (&qr,
+				     &SUFFIX(go_quad_one),
+				     &SUFFIX(go_quad_sqrt2));
+	else {
+		QUAD api;
+		SUFFIX(go_quad_mul) (&api, a, &SUFFIX(go_quad_pi));
+		SUFFIX(do_sin) (&qr, &api, k & 1);
+	}
+
+	if (k & 2) {
+		qr.h = 0 - qr.h;
+		qr.l = 0 - qr.l;
+	}
+
+	*res = qr;
+}
+
+/**
+ * go_quad_sin: (skip)
+ **/
+/**
+ * go_quad_sinl: (skip)
+ **/
+void
+SUFFIX(go_quad_sin) (QUAD *res, const QUAD *a)
+{
+	int k;
+	QUAD a0;
+
+	if (SUFFIX(reduce_pi_half) (&a0, a, &k))
+		SUFFIX(go_quad_init) (res, SUFFIX(sin) (a->h));
+	else
+		SUFFIX(do_sin) (res, &a0, k);
+}
+
+/**
+ * go_quad_sinpi: (skip)
+ **/
+/**
+ * go_quad_sinpil: (skip)
+ **/
+void
+SUFFIX(go_quad_sinpi) (QUAD *res, const QUAD *a)
+{
+	int k;
+	QUAD a0;
+
+	SUFFIX(reduce_half) (&a0, a, &k);
+	SUFFIX(do_sinpi) (res, &a0, k);
+}
+
+/**
+ * go_quad_asin: (skip)
+ **/
+/**
+ * go_quad_asinl: (skip)
+ **/
+void
+SUFFIX(go_quad_asin) (QUAD *res, const QUAD *a)
+{
+	QUAD aa, aam1;
+
+	aa.h = SUFFIX(fabs) (a->h);
+	aa.l = SUFFIX(fabs) (a->l);
+	SUFFIX(go_quad_sub) (&aam1, &aa, &SUFFIX(go_quad_one));
+	if (aam1.h > 0) {
+		SUFFIX(go_quad_init) (res, SUFFIX(go_nan));
+		return;
+	}
+
+	SUFFIX(go_quad_agm_internal) (res, AGM_ARCSIN, a);
+}
+
+/**
+ * go_quad_cos: (skip)
+ **/
+/**
+ * go_quad_cosl: (skip)
+ **/
+void
+SUFFIX(go_quad_cos) (QUAD *res, const QUAD *a)
+{
+	int k;
+	QUAD a0;
+
+	if (SUFFIX(reduce_pi_half) (&a0, a, &k))
+		SUFFIX(go_quad_init) (res, SUFFIX(cos) (a->h));
+	else
+		SUFFIX(do_sin) (res, &a0, k + 1);
+}
+
+/**
+ * go_quad_cospi: (skip)
+ **/
+/**
+ * go_quad_cospil: (skip)
+ **/
+void
+SUFFIX(go_quad_cospi) (QUAD *res, const QUAD *a)
+{
+	int k;
+	QUAD a0;
+
+	SUFFIX(reduce_half) (&a0, a, &k);
+	SUFFIX(do_sinpi) (res, &a0, k + 1);
+}
+
+/**
+ * go_quad_acos: (skip)
+ **/
+/**
+ * go_quad_acosl: (skip)
+ **/
+void
+SUFFIX(go_quad_acos) (QUAD *res, const QUAD *a)
+{
+	QUAD aa, aam1;
+
+	aa.h = SUFFIX(fabs) (a->h);
+	aa.l = SUFFIX(fabs) (a->l);
+	SUFFIX(go_quad_sub) (&aam1, &aa, &SUFFIX(go_quad_one));
+	if (aam1.h > 0) {
+		SUFFIX(go_quad_init) (res, SUFFIX(go_nan));
+		return;
+	}
+
+	SUFFIX(go_quad_agm_internal) (res, AGM_ARCCOS, &aa);
+	if (a->h < 0)
+		SUFFIX(go_quad_sub) (res, &SUFFIX(go_quad_pi), res);
+}
